package com.iot.modbus_rtcp.jobs;

import cn.hutool.core.map.MapUtil;
import com.iot.modbus_rtcp.config.EquipmentIPProperties;
import com.iot.modbus_rtcp.controller.ModbusTCPController;
import com.iot.modbus_rtcp.dto.CommandTypeComparable;
import com.iot.modbus_rtcp.dto.ModbusCommandDto;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.ObjectUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

/**
 * @author <a href="mailto:scwsl@foxmail.com">王仕龙</a>
 * 2024/11/23 18:33
 */
@Slf4j
@Component
@RequiredArgsConstructor
public class AutoCollectJobs {

    public static final String COMMAND_REF_TYPE_DEVICE = "DEVICE";
    public static final String COMMAND_REF_TYPE_PRODUCT_CODE = "DEVICE_PRODUCT_CODE";

    private static final String SQL = """
            SELECT * FROM (
                SELECT
                  c.id AS commandId,
                  d.id AS deviceId,
                  c.ref_type as refType,
                  c.command,
                  c.message_length messageLength,
                  d.gateway_sn AS identifier
                FROM
                  commands c JOIN device d ON c.ref_id = d.id
                WHERE
                  c.type = 'COLLECTION' AND c.ref_type = 'DEVICE'
                UNION ALL
                SELECT
                  c.id AS commandId,
                  d.id AS deviceId,
                  c.ref_type as refType,
                  c.command,
                  c.message_length messageLength,
                  d.gateway_sn AS identifier
                FROM
                  commands c JOIN device d ON c.ref_id = d.product
                WHERE
                  c.type = 'COLLECTION' AND c.ref_type = 'DEVICE_PRODUCT_CODE'
            ) a ORDER BY a.deviceId, a.commandId
            """;

    private final JdbcTemplate jdbcTemplate;
    private final ModbusTCPController controller;
    private final EquipmentIPProperties equipmentIPProperties;
    private final ThreadPoolExecutor collectThreadPool = new ThreadPoolExecutor(Math.min(10, Runtime.getRuntime().availableProcessors())
            , Math.min(100, Runtime.getRuntime().availableProcessors()), 60, TimeUnit.SECONDS,
            new ArrayBlockingQueue<>(5000), new ThreadFactory() {
        AtomicInteger index = new AtomicInteger(1);

        @Override
        public Thread newThread(Runnable runnable) {
            Thread thread = new Thread(runnable, "Modbus-auto-collect-thread-" + this.index.getAndIncrement());
            thread.setDaemon(true);
            thread.setPriority(Thread.MIN_PRIORITY);
            return thread;
        }
    }, new ThreadPoolExecutor.CallerRunsPolicy());

    @Scheduled(cron = "0/30 * * * * ? ")
    public void autoCollect() {
        int pageIndex = 0;
        int pageSize = 1000;
        long timestamp = System.currentTimeMillis();
        List<Map<String, Object>> resultList = null;
        while (Objects.isNull(resultList) || resultList.size() >= pageSize) {
            resultList = this.jdbcTemplate.queryForList(SQL + " LIMIT " + (pageIndex++ * pageSize) + ", " + pageSize);
            if (ObjectUtils.isEmpty(resultList)) {
                continue;
            }
            resultList.stream()
                    .filter(item -> {
                        String gatewaySn = MapUtil.getStr(item, "identifier");
                        String gatewayIdentifier = this.equipmentIPProperties.put(gatewaySn);
                        if (Objects.nonNull(ModbusTCPController.nettyServer.getGroup()
                                .get(gatewayIdentifier))) {
                            item.put("identifier", gatewayIdentifier);
                            return true;
                        }
                        log.warn("Gateway {} is disconnected and does not collect data", gatewaySn);
                        return false;
                    })
                    .collect(Collectors.groupingBy(item -> MapUtil.getStr(item, "deviceId")))
                    .forEach((deviceId, commandList) -> {
                        Map<String, List<Map<String, Object>>> refTypeCommandListMap = commandList.stream()
                                .collect(Collectors.groupingBy(item -> MapUtil.getStr(item, "refType")));
                        List<Map<String, Object>> deviceCommandList = refTypeCommandListMap.get(COMMAND_REF_TYPE_DEVICE);
                        if (ObjectUtils.isEmpty(deviceCommandList)) {
                            deviceCommandList = refTypeCommandListMap.get(COMMAND_REF_TYPE_PRODUCT_CODE);
                        }
                        if (ObjectUtils.isNotEmpty(deviceCommandList)) {
                            List<ModbusCommandDto> collectCommondList = deviceCommandList.stream()
                                    .map(item -> {
                                        String identifier = MapUtil.getStr(item, "identifier");
                                        return ModbusCommandDto.builder()
                                                .command(MapUtil.getStr(item, "command"))
                                                .identifier(identifier)
                                                .length(MapUtil.getInt(item, "messageLength"))
                                                .type(CommandTypeComparable.CommandType.COLLECTION)
                                                .key(StringUtils.joinWith("/", identifier,
                                                        MapUtil.getStr(item, "deviceId"),
                                                        MapUtil.getStr(item, "commandId"),
                                                        timestamp))
                                                .build();
                                    })
                                    .collect(Collectors.toList());
                            if (ObjectUtils.isNotEmpty(collectCommondList)) {
                                this.collectThreadPool.execute(() -> this.controller.collect(collectCommondList));
                            }
                        }
                    });
        }
    }
}
